----------------------------------------------------------------------
-- Polynomials Chapter 6 package body (polynomials.adb)
-- 
-- Defines useful funtions for polynomial operations
-- of Chapter 6 examples
---------------------------------------------------------------------
package body Polynomials is
   function Degree(A: Polynomial) return Natural is
      X: Natural;
   begin
      for I in 0 .. M loop if A(I) > 0 then X := I; end if; end loop;
      return X;
   end Degree; 
   function Subtract(A, B: Polynomial) return Polynomial is
      C: Polynomial;
   begin
      for I in 0 .. M loop C(I) := (A(I)-B(I)) mod P; end loop;
      return C;
   end Subtract;
   function Invert(Y: Natural_Mod_P) return Natural_Mod_P is
      A, B, C, D, Q, R, Next_D: Integer;
   begin
      A := P; B := Y; C := 0; D := 1;
      while B > 1 loop
         Q := A/B; R := A mod B; Next_D := C - D*Q;
         A := B; B := R; C := D; D := Next_D;
      end loop;
      return (D mod P);
   end Invert;
   function Product(A: Polynomial; B: Natural_Mod_P) return Polynomial is
      C: Polynomial;
   begin
      for I in 0 .. M loop C(I) := (A(I)*B) mod P; end loop;
      return C;
   end Product;
   function Shift_One(A: Polynomial) return Polynomial is
      B: Polynomial;
   begin
      for I in 0 .. M-1 loop B(I) := A(I+1); end loop;
      B(M) := 0;
      return B;
   end Shift_One;
   function Divide_By_X(A, F: Polynomial) return Polynomial is
   begin
      return Shift_One(Subtract(A, Product(F, ((A(0)*Invert(F(0))) mod P))));
   end Divide_By_X;
   function Add(A, B: Polynomial) return Polynomial is
      C: Polynomial;
   begin
      for I in 0 .. M loop C(I) := (A(I)+B(I)) mod P; end loop;
      return C;
   end Add;
   function Multiply_By_X(A, F: Polynomial) return Polynomial is
      C, D: Polynomial;
   begin
      for I in 1 .. M-1 loop C(I) := A(I-1);end loop;
      C(0) := 0; C(M) := 0;   
      for I in 0 .. M-1 loop D(I) := F(I);end loop;
      D(M) := 0;
      return Subtract(C, Product(D, A(M-1)));
   end Multiply_By_X;
   
   function Multiply_By_X(A: Polynomial) return Polynomial is
      C: Polynomial;
   begin
      for I in 1 .. M loop C(I) := A(I-1);end loop;
      C(0) := 0;    
      return C;
   end Multiply_By_X;

   function Product_Mod_F(A, B, F: Polynomial) return Polynomial is
      C: Polynomial;
   begin
      for I in 0 .. M loop C(I) := 0; end loop;
      for I in 0 .. M-1 loop
         C := Add(Multiply_By_X(C, F), Product(A,B(M-1-I)));
      end loop;
      return C;
   end Product_Mod_F;
   
   function Quotient(Num, Den: Polynomial) return Polynomial is
      R, Q, A, B: Polynomial;
      U, V: Natural;
   begin
      R := Num;
      for I in 0 .. M loop Q(I) := 0; end loop;
      U := Degree(Num); V := Degree(Den);
      while U >= V loop
         Q(U-V) := (R(U)*Invert(Den(V))) mod P;
         A := Product(Den, Q(U-V));
         for I in U-V .. M loop B(I) := A(I-U+V); end loop;
         for I in 0 .. U-V-1 loop B(I) := 0; end loop;
         R := Subtract(R, B);
         U := Degree(R);
      end loop;
      return Q;
   end Quotient;
   function Remainder(Num, Den: Polynomial) return Polynomial is
      R, Q, A, B: Polynomial;
      U, V: Natural;
   begin
      R := Num;
      for I in 0 .. M loop Q(I) := 0; end loop;
      U := Degree(Num); V := Degree(Den);
      while U >= V loop
         Q(U-V) := (R(U)*Invert(Den(V))) mod P;
         A := Product(Den, Q(U-V));
         for I in U-V .. M loop B(I) := A(I-U+V); end loop;
         for I in 0 .. U-V-1 loop B(I) := 0; end loop;
         R := Subtract(R, B);
         U := Degree(R);
      end loop;
      return R;
   end Remainder;
   function Shift(A: Polynomial; T: Natural) return Polynomial is
      Shifted: Polynomial;
   begin
      for I in T .. M loop Shifted(I) := A(I-T); end loop;
      for I in 0 .. T-1 loop Shifted(I) := 0; end loop;                  
      return Shifted;
   end Shift;
   function Shift(A, F: Polynomial; T: Natural) return Polynomial is
      Shifted: Polynomial;
   begin
      Shifted := A;
      for J in 1 .. T loop
         Shifted := Multiply_By_X(Shifted, F);
      end loop;                  
      return Shifted;
   end Shift;

   function Pseudo_Quotient(Num, Den: Polynomial) return Polynomial is
      Q: Polynomial;
      S, T: Natural;
   begin
      S := Degree(Num); T := Degree(Den);
      for I in 0 .. M loop Q(I) := 0; end loop;
      Q(S-T) := (Num(S)*Invert(Den(T))) mod P;
      return Q;
   end Pseudo_Quotient;

   function Pseudo_Remainder(Num, Den: Polynomial) return Polynomial is
      Q,R, Shifted_Den, Den_By_Q: Polynomial;
      S, T: Natural;
   begin
      S := Degree(Num); T := Degree(Den);
      for I in 0 .. M loop Q(I) := 0; end loop;
      Q(S-T) := (Num(S)*Invert(Den(T))) mod P;
      Shifted_Den := Shift(Den, S-T);   
      Den_By_Q := Product(Shifted_Den, Q(S-T));   
      R := Subtract(Num, Den_By_Q);
      return R;
   end Pseudo_Remainder;
   
   procedure Swap(A, B: in out Polynomial) is
      Old_A: Polynomial;
   begin
      Old_A := A; A := B; B := Old_A;
   end Swap;
   procedure Swap(S, T: in out Natural) is 
      Old_S: Natural;
   begin
      Old_S := S; S := T; T := Old_S;
   end Swap;

   --procedure pseudo_remainder(n_a: in polynomial; deg_a: in natural; 
   --   n_b: in polynomial; deg_b: in natural; 
   --   n_r: out polynomial; deg_r: out natural) 
   --is
   --   r: polynomial;
   --   d_r: natural;
   --begin   
   --   r := subtract(n_a, product(n_b,((invert(n_b(m))*n_a(m)) mod p)));
   --   d_r := deg_a;
   --   while r(m) = 0 loop r := multiply_by_x(r); d_r := d_r-1; end loop;
   --   n_r := r; deg_r :=d_r;
   --end pseudo_remainder;

   --procedure pseudo_remainder(m_a: in polynomial; deg_a: in natural; coef_a: in natural_mod_p;
   --   m_b: in polynomial; deg_b: in natural; coef_b: in natural_mod_p;
   --   m_r: out polynomial; deg_r: out natural; coef_r: out natural_mod_p) 
   --is
   --   r: polynomial;
   --   d_r: natural;
   --begin
   --   r := subtract(m_a, m_b);
   --   d_r := deg_a;
   --while r(m) = 0 loop r := multiply_by_x(r); d_r := d_r-1; end loop;
   --coef_r := (r(m) * coef_a) mod p; m_r := product(r, invert(r(m))); deg_r := d_r;
   --end pseudo_remainder;

   function Frobenius(J, I: Natural)return Natural_Mod_P is 
      C: constant Natural := 2;
      T: Natural;   
      B, Z: Natural_Mod_P;
   begin
      T := P/M;
      B := (C**T) mod P;
      Z := 1;
      for K in 1 .. I*J loop Z := (Z*B)mod P; end loop;
      return Z;
   end Frobenius;      
end Polynomials;

