----------------------------------------------------------------------------
-- Pseudo Euclidean Divider (pseudo_Euclidean_Divider.vhd)
--
-- Defines 7 entities:
--    - mod_239_inverter (read the table and gives the mod 239 inversion)
--    - mod_239_reducer (as in ch3) 
--    - mod_239_multiplier (as in ch3) 
--    - subtractor parameters (a package for subtractor) 
--    - subtractor (defines a modular subtraction) 
--    - pseudo_Euclidean_divider_parameters (a package)
--    - pseudo_Euclidean_divider (the divider)
--
----------------------------------------------------------------------------

------------------------------------------------------------
-- mod 239 inverter
------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity mod_239_inverter is
port (
  clk: in std_logic; 
  x: in std_logic_vector(7 downto 0);
  z: out std_logic_vector(7 downto 0)
);
end mod_239_inverter;

architecture table of mod_239_inverter is
  type RomType is array (0 to 238) of natural;
  constant ROM: RomType  := (0, 1, 120, 80, 60, 48, 40, 205, 30, 186, 24, 87, 20, 92, 222, 16,
    15, 225, 93, 151, 12, 148, 163, 52, 10, 153, 46, 62, 111, 33, 8, 54,
    127, 29, 232, 41, 166, 84, 195, 190, 6, 35, 74, 189, 201, 85, 26, 178,
    5, 200, 196, 75, 23, 230, 31, 113, 175, 130, 136, 158, 4, 192, 27, 129,
    183, 114, 134, 132, 116, 97, 140, 101, 83, 203, 42, 51, 217, 149, 95, 118,
    3, 180, 137, 72, 37, 45, 214, 11, 220, 94, 162, 218, 13, 18, 89, 78,
    122, 69, 100, 169, 98, 71, 157, 181, 131, 173, 115, 172, 135, 182, 176, 28,
    207, 55, 65, 106, 68, 143, 79, 237, 2, 160, 96, 171, 133, 174, 184, 32,
    211, 63, 57, 104, 67, 124, 66, 108, 58, 82, 168, 141, 70, 139, 170, 117,
    161, 150, 221, 226, 21, 77, 145, 19, 228, 25, 194, 202, 167, 102, 59, 236,
    121, 144, 90, 22, 188, 197, 36, 156, 138, 99, 142, 123, 107, 105, 125, 56,
    110, 212, 47, 235, 81, 103, 109, 64, 126, 208, 9, 216, 164, 43, 39, 234,
    61, 213, 154, 38, 50, 165, 204, 233, 49, 44, 155, 73, 198, 7, 210, 112,
    185, 231, 206, 128, 177, 193, 86, 229, 187, 76, 91, 227, 88, 146, 14, 224,
    223, 17, 147, 219, 152, 215, 53, 209, 34, 199, 191, 179, 159, 119, 238 );
begin
  z <= conv_std_logic_vector(ROM(conv_integer(x)),8);
end table;

------------------------------------------------------------
-- mod 239 reducer
------------------------------------------------------------
library ieee; use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
entity mod_239_reducer is
port (
  x: in std_logic_vector(15 downto 0);
  z: out std_logic_vector(7 downto 0)
  );
end mod_239_reducer;

architecture circuit of mod_239_reducer is
  signal x1_by_17, x0, sum, xx1_by_17, xx0, xxx: std_logic_vector(8 downto 0);
  signal long_xxx, minus_239, dif: std_logic_vector(9 downto 0);
  signal x2_by_33, long_sum, xx: std_logic_vector(9 downto 0);
begin
  x1_by_17 <= '0'&x(11 downto 8)&x(11 downto 8);
  x0 <= '0'&x(7 downto 0);
  sum <= x1_by_17 + x0;
  x2_by_33 <= '0'&x(15 downto 12)&'0'&x(15 downto 12);
  long_sum <= '0'&sum;
  xx <= x2_by_33 + long_sum;
  xx1_by_17 <= "000"&xx(9 downto 8) &"00"&xx(9 downto 8);
  xx0 <= '0'&xx(7 downto 0);
  xxx <= xx1_by_17 + xx0;
  minus_239 <= conv_std_logic_vector(273, 10);
  long_xxx <= '0'&xxx;
  dif <= long_xxx + minus_239;
  with dif(9) select z <= dif(7 downto 0) when '1', xxx(7 downto 0) when others;
end circuit;

------------------------------------------------------------
-- mod 239 multiplier
------------------------------------------------------------
library ieee; use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
entity mod_239_multiplier is
port (
  x, y: in std_logic_vector(7 downto 0);
  z: out std_logic_vector(7 downto 0)
);
end mod_239_multiplier;

architecture circuit of mod_239_multiplier is
  component mod_239_reducer is
  port 
  (x: in std_logic_vector(15 downto 0);
  z: out std_logic_vector(7 downto 0)
  );
  end component;
  signal product: std_logic_vector(15 downto 0);
begin
  product <= x*y;
  main_component: mod_239_reducer port map(product, z);
end circuit;

------------------------------------------------------------
-- subtractor parameters
------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package subtractor_parameters is
  constant k: natural := 8;
  constant p: std_logic_vector(k-1 downto 0) := conv_std_logic_vector(239, k);
end subtractor_parameters;

------------------------------------------------------------
-- subtractor
------------------------------------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.subtractor_parameters.all;
entity subtractor is
port (
  x, y: in std_logic_vector(k-1 downto 0);
  z: out std_logic_vector(k-1 downto 0)
);
end subtractor;

architecture rtl of subtractor is
  signal long_x, not_y, sum1: std_logic_vector(k downto 0);
  signal c1: std_logic;
  signal z1, z2: std_logic_vector(k-1 downto 0);
begin
  long_x <= '0'&x;
  not_gates: for i in 0 to k-1 generate not_y(i) <= not(y(i)); end generate;
  not_y(k) <= '0';
  sum1 <= 1 + long_x + not_y;
  c1 <= sum1(k);
  z1 <= sum1(k-1 downto 0);
  z2 <= z1 + p;
  with c1 select z <= z1 when '1', z2 when others;
end rtl;

------------------------------------------------------------
-- Pseudo Euclidean divider parameters
------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package pseudo_Euclidean_divider_parameters is
  constant k: natural := 8;
  constant p: std_logic_vector(k-1 downto 0) := conv_std_logic_vector(239, k);
  --subtype coefficient is natural range 0 to 238;
  constant m: natural := 17;
  --subtype degree is natural range 0 to m;
  type long_polynomial is array(m downto 0) of std_logic_vector(k-1 downto 0);
  type polynomial is array(m-1 downto 0) of std_logic_vector(k-1 downto 0);
  constant logm: natural := 5;
  constant f: long_polynomial :=
  ("00000001", "00000000", "00000000", "00101010", "00000000", "00000000", "00000000",
   "00000000", "00000000", "00000000", "00011101", "00000000", "00000000", "00000000", 
   "00000000", "00000000", "00000000", "11101101" );
  constant zero_poly: polynomial := (others => "00000000");
end pseudo_Euclidean_divider_parameters;

------------------------------------------------------------
-- pseudo Euclidean Divider
------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.pseudo_Euclidean_divider_parameters.all;
entity pseudo_Euclidean_divider is
port(
  g, h: in polynomial;
  clk, reset, start: in std_logic;  
  z: out polynomial;
  done: out std_logic
);
end pseudo_Euclidean_divider;

architecture circuit of pseudo_Euclidean_divider is
  component mod_239_inverter is
  port (
    clk: in std_logic; 
    x: in std_logic_vector(7 downto 0);
    z: out std_logic_vector(7 downto 0) );
  end component;
  component subtractor is
  port (
    x, y: in std_logic_vector(k-1 downto 0);
    z: out std_logic_vector(k-1 downto 0));
  end component;
  component mod_239_multiplier is
  port (
    x, y: in std_logic_vector(7 downto 0);
    z: out std_logic_vector(7 downto 0) );
  end component;

  signal n_a, n_b, n_r, next_a, next_b, next_r,
  sub1, sub2, out1, nbe, nr_by_x, nb_by_x, h_by_x,
  long_c, long_e: long_polynomial;
  signal c, d, e, next_c, next_e,
  sub3, sub4, out2, short_out1: polynomial;
  signal deg_a, deg_b, deg_r, dif, 
  next_da, next_db, next_dr,
  dr_minus1, db_minus1, count: std_logic_vector(logm-1 downto 0);
  signal inv_out, coef: std_logic_vector(k-1 downto 0);
  signal ce_a, ce_b, ce_d, ce_r, ce_e, sel_a, sel_b, sel_r, sel_sub,
  load, update, count_zero, deg_zero, swap, first, b_zero, r_zero: std_logic;
  signal sel_e: std_logic_vector(1 downto 0);

  type states is range 0 to 9;
  signal current_state: states;

begin

  long_c(m) <= "00000000"; long_e(m) <= "00000000";
  definition: for i in 0 to m-1 generate 
    long_c(i) <= c(i); long_e(i) <= e(i); end generate;
    with sel_sub select sub1 <= n_a when '0', long_c when others;
    with sel_sub select nbe <= n_b when '0', long_e when others;
    functions1: for i in 0 to m generate
    m1: mod_239_multiplier port map(coef, nbe(i), sub2(i));
    s1: subtractor port map(sub1(i), sub2(i), out1(i));
  end generate;
  by_x1: for i in 1 to m-1 generate sub3(i) <= e(i-1); end generate;
    sub3(0) <= "00000000";
    functions2: for i in 0 to m-1 generate
    m2: mod_239_multiplier port map(e(m-1), f(i), sub4(i));
    s2: subtractor port map(sub3(i), sub4(i), out2(i));
  end generate;
  dr_minus1 <= deg_r - 1;
  db_minus1 <= deg_b - 1;
  dif <= deg_a - deg_b;
  inverter: mod_239_inverter port map(clk, n_b(m), inv_out);
  m3: mod_239_multiplier port map(n_a(m), inv_out, coef);
  functions3: for i in 0 to m-1 generate
    m4: mod_239_multiplier port map(inv_out, d(i), z(i));
  end generate;
  by_x2: for i in 1 to m generate nr_by_x(i) <= n_r(i-1); end generate;
  nr_by_x(0) <= "00000000";
  with sel_r select next_r <= out1 when '0', nr_by_x when others;
  with sel_r select next_dr <= deg_a when '0', dr_minus1 when others;
  definition2: for i in 0 to m-1 generate short_out1(i) <= out1(i); end generate;
  with sel_e select next_e <= d when "00", out2 when "01", short_out1 when others;
  with sel_a select next_a <= n_b when '0', n_r when others;
  with sel_a select next_da <= deg_b when '0', deg_r when others;
  by_x3: for i in 1 to m generate nb_by_x(i) <= n_b(i-1); end generate;
  nb_by_x(0) <= "00000000";
  with sel_b select next_b <= nb_by_x when '0', n_r when others;
  with sel_b select next_db <= db_minus1 when '0', deg_r when others;
  with sel_a select next_c <= d when '0', e when others;

  registers_ac: process(clk)
  begin
  if clk'event and clk = '1' then
    if first = '1' then n_a <= f; deg_a <= conv_std_logic_vector(m, logm); c <= zero_poly;
    elsif ce_a = '1' then n_a <= next_a; deg_a <= next_da; c <= next_c; 
    end if;
  end if;
  end process registers_ac;

  by_x4: for i in 1 to m generate h_by_x(i) <= h(i-1); end generate;
  h_by_x(0) <= "00000000";
  
  register_b: process(clk)
  begin
  if clk'event and clk = '1' then
    if first = '1' then n_b <= h_by_x; deg_b <= conv_std_logic_vector(m-1, logm);
    elsif ce_b = '1' then n_b <= next_b; deg_b <= next_db; 
    end if;
  end if;
  end process register_b;

  register_r: process(clk)
  begin
  if clk'event and clk = '1' then
    if ce_r = '1' then n_r <= next_r; deg_r <= next_dr; end if;
  end if;
  end process register_r;

  register_d: process(clk)
  begin
  if clk'event and clk = '1' then
    if first = '1' then d <= g;
    elsif ce_d = '1' then d <= e; 
    end if;
  end if;
  end process register_d;

  register_e: process(clk)
  begin
  if clk'event and clk = '1' then
    if ce_e = '1' then e <= next_e; 
    end if;
  end if;
  end process register_e;

  swap <= '1' when deg_b < deg_r else '0';
  deg_zero <= '1' when deg_b <= conv_std_logic_vector(0, logm) else '0';
  b_zero <= '1' when n_b(m) <= conv_std_logic_vector(0, k) else '0';
  r_zero <= '1' when n_r(m) <= conv_std_logic_vector(0, k) else '0';

  counter: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then count <= dif;
    elsif update = '1' then count <= count - 1;
    end if;
  end if;
  end process counter;
  count_zero <= '1' when count = conv_std_logic_vector(0, logm) else '0';

  control_unit: process(clk, reset, current_state, count_zero, b_zero, r_zero)
  begin
  case current_state is
    when 0 to 1 => first <= '0'; sel_a <= '0'; ce_a <= '0'; sel_b <= '0'; ce_b <= '0'; ce_d <= '0';
              sel_r <= '0'; sel_sub <= '0'; ce_r <= '0'; sel_e <= "00"; ce_e <= '0';
              load <= '0'; update <= '0'; done <= '1';
    when 2 => first <= '1'; sel_a <= '0'; ce_a <= '0'; sel_b <= '0'; ce_b <= '0'; ce_d <= '0';
              sel_r <= '0'; sel_sub <= '0'; ce_r <= '0'; sel_e <= "00"; ce_e <= '0';
              load <= '0'; update <= '0'; done <= '0';
    when 3 => first <= '0'; sel_a <= '0'; ce_a <= '0'; sel_b <= '0'; 
              if b_zero = '1' then ce_b <= '1'; else ce_b <= '0'; end if; 
              ce_d <= '0'; sel_r <= '0'; sel_sub <= '0'; ce_r <= '0'; sel_e <= "00"; ce_e <= '0';
              load <= '0'; update <= '0'; done <= '0';
    when 4 => first <= '0'; sel_a <= '0'; ce_a <= '0'; sel_b <= '0'; ce_b <= '0'; ce_d <= '0';
              sel_r <= '0'; sel_sub <= '0'; sel_e <= "00"; 
              if deg_zero = '1' then  ce_r <= '0'; ce_e <= '0'; load <= '0';
              else ce_r <= '1'; ce_e <= '1'; load <= '1';
              end if;
              update <= '0'; done <= '0';
    when 5 => first <= '0'; sel_a <= '0'; ce_a <= '0'; sel_b <= '0'; ce_b <= '0'; ce_d <= '0';
              if r_zero = '1' and count_zero = '0' then 
                sel_r <= '1'; sel_sub <= '0'; ce_r <= '1'; sel_e <= "01"; ce_e <= '1'; update <= '1'; 
              elsif r_zero = '1' and count_zero = '1' then 
                sel_r <= '1'; sel_sub <= '0'; ce_r <= '1'; sel_e <= "00"; ce_e <= '0'; update <= '0'; 
              elsif r_zero = '0' and count_zero = '0' then 
                sel_r <= '0'; sel_sub <= '0'; ce_r <= '0'; sel_e <= "01"; ce_e <= '1'; update <= '1'; 
              else  
                sel_r <= '0'; sel_sub <= '1'; ce_r <= '0'; sel_e <= "10"; ce_e <= '1'; update <= '0';
              end if;
              load <= '0'; done <= '0';

    when 6 => first <= '0'; sel_a <= '0'; ce_a <= '0'; sel_b <= '0'; ce_b <= '0'; ce_d <= '0';
              if r_zero = '1' then 
                sel_r <= '1'; sel_sub <= '0'; ce_r <= '1'; sel_e <= "00"; ce_e <= '0'; update <= '0'; 
              else  
                sel_r <= '0'; sel_sub <= '0'; ce_r <= '0'; sel_e <= "00"; ce_e <= '0'; update <= '0';
              end if;
              load <= '0'; done <= '0';
    when 7 => first <= '0'; sel_a <= '0'; ce_a <= '0'; sel_b <= '0'; ce_b <= '0'; ce_d <= '0';
              if count_zero = '0' then 
                sel_r <= '0'; sel_sub <= '0'; ce_r <= '0'; sel_e <= "01"; ce_e <= '1'; update <= '1'; 
              else  
                sel_r <= '0'; sel_sub <= '0'; ce_r <= '0'; sel_e <= "00"; ce_e <= '0'; update <= '0';
              end if;
              load <= '0'; done <= '0';
    when 8 => first <= '0'; sel_a <= '0'; ce_a <= '0'; sel_b <= '0'; ce_b <= '0'; ce_d <= '0';
              sel_r <= '0'; sel_sub <= '1'; ce_r <= '0'; sel_e <= "10"; ce_e <= '1';
              load <= '0'; update <= '0'; done <= '0';
    when 9 => first <= '0'; 
              if swap = '0' then sel_a <= '0'; ce_a <= '1'; sel_b <= '1'; ce_b <= '1'; ce_d <= '1';
              else sel_a <= '1'; ce_a <= '1'; sel_b <= '0'; ce_b <= '0'; ce_d <= '0';
              end if;
              sel_r <= '0'; sel_sub <= '0'; ce_r <= '0'; sel_e <= "00"; ce_e <= '0';
              load <= '1'; update <= '0'; done <= '0';
  end case;

  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => if b_zero = '0' then current_state <= 4; end if;
      when 4 => if deg_zero = '0' then current_state <= 5; else current_state <= 0; end if;
      when 5 => if r_zero = '1' and count_zero = '0' then current_state <= 5; 
                elsif r_zero = '1' and count_zero = '1' then current_state <= 6; 
                elsif r_zero = '0' and count_zero ='0'then current_state <= 7;
                else current_state <= 9;
                end if;
      when 6 => if r_zero = '0' then current_state <= 8; end if;
      when 7 => if count_zero = '1' then current_state <= 8; end if;
      when 8 => current_state <= 9;
      when 9 => current_state <= 4; 
    end case;
  end if;
  end process control_unit;

end circuit;


 