----------------------------------------------------------------------------
-- reduction to multiplication (reduction_to_multiplication.vhd)
--
-- Defines the following entities:
--    - reduction_to_multiplications_parameters (a package)
--    - mod_239_inverter (read the table and gives the mod 239 inversion)
--    - mod_239_reducer (as in ch3) 
--    - mod_239_multiplier (as in ch3) 
--    - subtractor mod P
--    - mof_f_multiplier (LSE-first polynomial multiplier)
--    - reduction_to_multiplications (the divider)
--
----------------------------------------------------------------------------
 
------------------------------------------------------------
-- reduction_to_multiplications_parameters
------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package reduction_to_multiplications_parameters is
  constant k: natural := 8;
  constant p: std_logic_vector(k-1 downto 0) := conv_std_logic_vector(239, k);
  --subtype coefficient is natural range 0 to 238;
  constant m: natural := 17;
  --subtype degree is natural range 0 to m;
  type long_polynomial is array(m downto 0) of std_logic_vector(k-1 downto 0);
  type polynomial is array(m-1 downto 0) of std_logic_vector(k-1 downto 0);
  constant logm: natural := 6;
  constant zero_coef: std_logic_vector(k-1 downto 0) := (others => '0');
  constant one_coef: std_logic_vector(k-1 downto 0) := (0 => '1', others => '0');
  constant f: long_polynomial := 
    ("00000001", "00000000", "00000000", "00000000", "00000000", "00000000", "00000000", 
     "00000000", "00000000", "00000000", "00000000", "00000000", "00000000", "00000000", 
     "00000000", "00000000", "00000000", "11101101" );
  constant one_poly: polynomial := ( 0 => one_coef, others => zero_coef);
  constant s: natural := 128;
  constant logs: natural := 7;
  constant r: std_logic_vector(s-1 downto 0) := x"559F5BD6ED9A760F32B6B45FE07B7880"; --239^17-1
  constant zero_poly: polynomial := (others => zero_coef);

end reduction_to_multiplications_parameters;

------------------------------------------------------------
-- mod 239 inverter
------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity mod_239_inverter is
port (
  clk: in std_logic; 
  x: in std_logic_vector(7 downto 0);
  z: out std_logic_vector(7 downto 0)
);
end mod_239_inverter;

architecture table of mod_239_inverter is
  type RomType is array (0 to 238) of natural;
  constant ROM: RomType  := (0, 1, 120, 80, 60, 48, 40, 205, 30, 186, 24, 87, 20, 92, 222, 16,
    15, 225, 93, 151, 12, 148, 163, 52, 10, 153, 46, 62, 111, 33, 8, 54,
    127, 29, 232, 41, 166, 84, 195, 190, 6, 35, 74, 189, 201, 85, 26, 178,
    5, 200, 196, 75, 23, 230, 31, 113, 175, 130, 136, 158, 4, 192, 27, 129,
    183, 114, 134, 132, 116, 97, 140, 101, 83, 203, 42, 51, 217, 149, 95, 118,
    3, 180, 137, 72, 37, 45, 214, 11, 220, 94, 162, 218, 13, 18, 89, 78,
    122, 69, 100, 169, 98, 71, 157, 181, 131, 173, 115, 172, 135, 182, 176, 28,
    207, 55, 65, 106, 68, 143, 79, 237, 2, 160, 96, 171, 133, 174, 184, 32,
    211, 63, 57, 104, 67, 124, 66, 108, 58, 82, 168, 141, 70, 139, 170, 117,
    161, 150, 221, 226, 21, 77, 145, 19, 228, 25, 194, 202, 167, 102, 59, 236,
    121, 144, 90, 22, 188, 197, 36, 156, 138, 99, 142, 123, 107, 105, 125, 56,
    110, 212, 47, 235, 81, 103, 109, 64, 126, 208, 9, 216, 164, 43, 39, 234,
    61, 213, 154, 38, 50, 165, 204, 233, 49, 44, 155, 73, 198, 7, 210, 112,
    185, 231, 206, 128, 177, 193, 86, 229, 187, 76, 91, 227, 88, 146, 14, 224,
    223, 17, 147, 219, 152, 215, 53, 209, 34, 199, 191, 179, 159, 119, 238 );
begin
  z <= conv_std_logic_vector(ROM(conv_integer(x)),8);
end table;

------------------------------------------------------------
-- mod 239 reducer
------------------------------------------------------------
library ieee; use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
entity mod_239_reducer is
port (
  x: in std_logic_vector(15 downto 0);
  z: out std_logic_vector(7 downto 0)
  );
end mod_239_reducer;

architecture circuit of mod_239_reducer is
  signal x1_by_17, x0, sum, xx1_by_17, xx0, xxx: std_logic_vector(8 downto 0);
  signal long_xxx, minus_239, dif: std_logic_vector(9 downto 0);
  signal x2_by_33, long_sum, xx: std_logic_vector(9 downto 0);
begin
  x1_by_17 <= '0'&x(11 downto 8)&x(11 downto 8);
  x0 <= '0'&x(7 downto 0);
  sum <= x1_by_17 + x0;
  x2_by_33 <= '0'&x(15 downto 12)&'0'&x(15 downto 12);
  long_sum <= '0'&sum;
  xx <= x2_by_33 + long_sum;
  xx1_by_17 <= "000"&xx(9 downto 8) &"00"&xx(9 downto 8);
  xx0 <= '0'&xx(7 downto 0);
  xxx <= xx1_by_17 + xx0;
  minus_239 <= conv_std_logic_vector(273, 10);
  long_xxx <= '0'&xxx;
  dif <= long_xxx + minus_239;
  with dif(9) select z <= dif(7 downto 0) when '1', xxx(7 downto 0) when others;
end circuit;

------------------------------------------------------------
-- mod 239 multiplier
------------------------------------------------------------
library ieee; use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
entity mod_239_multiplier is
port (
  x, y: in std_logic_vector(7 downto 0);
  z: out std_logic_vector(7 downto 0)
);
end mod_239_multiplier;

architecture circuit of mod_239_multiplier is
  component mod_239_reducer is
  port (
    x: in std_logic_vector(15 downto 0);
    z: out std_logic_vector(7 downto 0) );
  end component;
  signal product: std_logic_vector(15 downto 0);
begin
  product <= x*y;
  main_component: mod_239_reducer port map(product, z);
end circuit;

------------------------------------------------------------
-- subtractor mod P 
------------------------------------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.reduction_to_multiplications_parameters.all;

entity subtractor_mod_P is
port (
  x, y: in std_logic_vector(K-1 downto 0);
  z: out std_logic_vector(K-1 downto 0)
  );
end subtractor_mod_P;

architecture rtl of subtractor_mod_P is
  signal z1, z2: std_logic_vector(K downto 0);
begin
  z1 <= ('0' & x) - y;
  z2 <= z1 + P;
  with z1(K) select 
       z <= z1(K-1 downto 0) when '0', 
            z2(K-1 downto 0) when others;
end rtl;
------------------------------------------------------------
-- mod f multiplier 
------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.reduction_to_multiplications_parameters.all;

entity LSE_first_mod_f_mult is
port(
  a, b: in polynomial;
  clk, reset, start: in std_logic;  
  z: out polynomial;
  done: out std_logic
  );
end LSE_first_mod_f_mult;

architecture circuit of LSE_first_mod_f_mult is
  component subtractor_mod_P is
  port (
    x, y: in std_logic_vector(K-1 downto 0);
    z: out std_logic_vector(K-1 downto 0) );
  end component;
  component mod_239_reducer is
  port(
    x: in std_logic_vector(15 downto 0);
    z: out std_logic_vector(7 downto 0) );
  end component;

  constant zero_coef: std_logic_vector(K-1 downto 0) := (others => '0');
  signal int_a, int_b, next_a, next_c, c, mult_sub: polynomial;
  type polynomial_2K is array(M-1 downto 0) of std_logic_vector(2*K-1 downto 0);
  signal mult_add, mult_f_x_a: polynomial_2K;
  constant ZERO: std_logic_vector(K-1 downto 0) := (others => '0'); 

  signal load, update, count_equal_zero: std_logic;
  signal count: std_logic_vector(logM-1 downto 0);
  type states is range 0 to 3;
  signal current_state: states;

begin

  next_c_calc: for i in 0 to m-1 generate
    mult_add(i) <= ( int_b(0) * int_a(i) ) + c(i);
    comp1: mod_239_reducer port map(mult_add(i), next_c(i));
  end generate;
  
  next_a_calc: for i in 1 to m-1 generate
    mult_f_x_a(i) <= F(i) * int_a(M-1);
    comp1: mod_239_reducer port map (mult_f_x_a(i), mult_sub(i));
    comp2: subtractor_mod_P port map (int_a(i-1), mult_sub(i), next_a(i));
  end generate;
  mult_f_x_a(0) <= ( F(0) * int_a(M-1) );
  comp1: mod_239_reducer port map(mult_f_x_a(0), mult_sub(0));
  comp2: subtractor_mod_P port map(ZERO, mult_sub(0), next_a(0));

  registers_abc: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      c <= ZERO_POLY;
      int_b <= b;
      int_a <= a;
    elsif update = '1' then 
      c <= next_c;
      int_b <= zero_coef & int_b(M-1 downto 1);
      int_a <= next_a;
    end if;
  end if;
  end process registers_abc;

  z <= c;

  counter: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then count <= conv_std_logic_vector(M-1, LOGM);
    elsif update = '1' then count <= count - 1;
    end if;
  end if;
  end process counter;

  count_equal_zero <= '1' when count = 0 else '0';

  control_unit: process(clk, reset, count_equal_zero, current_state)
  begin
  case current_state is
    when 0 to 1 => load <= '0'; update <= '0'; done <= '1';
    when 2 => load <= '1'; update <= '0'; done <= '0';
    when 3 => load <= '0'; update <= '1'; done <= '0';
  end case;

  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => if count_equal_zero = '1' then current_state <= 0; end if;
    end case;
  end if;

  end process control_unit;

end circuit;
------------------------------------------------------------
-- reduction_to_multiplications
------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.reduction_to_multiplications_parameters.all;
entity reduction_to_multiplications is
port(
  g, h: in polynomial;
  clk, reset, start: in std_logic;  
  z: out polynomial;
  done: out std_logic
  );
end reduction_to_multiplications;

architecture circuit of reduction_to_multiplications is
  component mod_239_inverter is
  port (
    clk: in std_logic; 
    x: in std_logic_vector(7 downto 0);
    z: out std_logic_vector(7 downto 0) );
  end component;
  component mod_239_multiplier is
  port (
  x, y: in std_logic_vector(7 downto 0);
  z: out std_logic_vector(7 downto 0));
  end component;
--  ------------------------------------
--  --provisional
--  ------------------------------------
--  component mod_f_multiplier is
--  port(
--    a, b: in polynomial;
--    clk, reset, start: in std_logic;  
--    z: out polynomial;
--    done: out std_logic );
--  end component;
--  ------------------------------------  
  component LSE_first_mod_f_mult is
  port(
    a, b: in polynomial;
    clk, reset, start: in std_logic;  
    z: out polynomial;
    done: out std_logic );
  end component;  

  signal mult_in2, mult_out1, next_ea, e, mult_out2: polynomial;
  signal a0, inv_a0: std_logic_vector(k-1 downto 0);
  signal sel_e, ce_e, ce_a, start_mult, mult_done, load, count_equal_zero, update, serial_out: std_logic;
  signal sel_mult: std_logic_vector(1 downto 0);
  signal count: std_logic_vector(logs-1 downto 0);
  signal int_r: std_logic_vector(s-1 downto 0);
  type states is range 0 to 15;
  signal current_state: states;
begin

  with sel_mult select mult_in2 <= e when "00", h when "01", g when others;
  a_mod_f_multiplier: LSE_first_mod_f_mult port map(e, mult_in2, clk, reset, start_mult, mult_out1, mult_done);
  with sel_e select next_ea <= mult_out1 when '0', mult_out2 when others;
  an_inverter: mod_239_inverter port map(clk, a0, inv_a0);
  
  mod_p_multipliers: for i in 0 to m-1 generate
    a_mod_p_multiplier: mod_239_multiplier port map(e(i), inv_a0, mult_out2(i));
  end generate;

  register_e: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then e <= one_poly;
    elsif ce_e = '1' then e <= next_ea;
    end if;
  end if;
  end process;
  z <= e;

  register_a: process(clk)
  begin
  if clk'event and clk = '1' then
    if ce_a = '1' then a0 <= next_ea(0);
    end if;
  end if;
  end process;

  a_counter: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then count <= conv_std_logic_vector(s-1, logs);
    elsif update = '1' then count <= count - 1;
    end if;
  end if;
  end process;
  count_equal_zero <= '1' when count = conv_std_logic_vector(0, logs) else '0';

  a_shift_register: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then int_r <= r;
    elsif update = '1' then
      for i in s-1 downto 1 loop int_r(i) <= int_r(i-1); end loop;
      int_r(0) <= '0';
    end if;
  end if;
  end process;
  serial_out <= int_r(s-1);

  control_unit: process(clk, reset, current_state, mult_done, count_equal_zero)
  begin
  case current_state is
    when 0 to 1 => sel_mult <= "00"; sel_e <= '0'; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; update <= '0'; done <= '1';
    when 2 => sel_mult <= "00"; sel_e <= '0'; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '1'; update <= '0'; done <= '0';
    when 3 => sel_mult <= "00"; sel_e <= '0'; ce_e <= '0'; ce_a <= '0'; start_mult <= '1'; load <= '0'; update <= '1'; done <= '0';
    when 4 => sel_mult <= "00"; sel_e <= '0'; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; update <= '0'; done <= '0';
    when 5 => sel_mult <= "00"; sel_e <= '0'; ce_e <= '1'; ce_a <= '0'; start_mult <= '0'; load <= '0'; update <= '0'; done <= '0';
    when 6 => sel_mult <= "01"; sel_e <= '0'; ce_e <= '0'; ce_a <= '0'; start_mult <= '1'; load <= '0'; update <= '0'; done <= '0';
    when 7 => sel_mult <= "01"; sel_e <= '0'; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; update <= '0'; done <= '0';
    when 8 => sel_mult <= "01"; sel_e <= '0'; ce_e <= '1'; ce_a <= '0'; start_mult <= '0'; load <= '0'; update <= '0'; done <= '0';
    when 9 => sel_mult <= "01"; sel_e <= '0'; ce_e <= '0'; ce_a <= '0'; start_mult <= '1'; load <= '0'; update <= '0'; done <= '0';
    when 10 => sel_mult <= "01"; sel_e <= '0'; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; update <= '0'; done <= '0';
    when 11 => sel_mult <= "01"; sel_e <= '0'; ce_e <= '0'; ce_a <= '1'; start_mult <= '0'; load <= '0'; update <= '0'; done <= '0';
    when 12 => sel_mult <= "00"; sel_e <= '1'; ce_e <= '1'; ce_a <= '0'; start_mult <= '0'; load <= '0'; update <= '0'; done <= '0';
    when 13 => sel_mult <= "10"; sel_e <= '0'; ce_e <= '0'; ce_a <= '0'; start_mult <= '1'; load <= '0'; update <= '0'; done <= '0';
    when 14 => sel_mult <= "10"; sel_e <= '0'; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; update <= '0'; done <= '0';
    when 15 => sel_mult <= "10"; sel_e <= '0'; ce_e <= '1'; ce_a <= '0'; start_mult <= '0'; load <= '0'; update <= '0'; done <= '0';
  end case;

  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => current_state <= 4;
      when 4 => if mult_done = '1' then current_state <= 5; end if;
      when 5 => if serial_out = '1' then current_state <= 6;
                elsif count_equal_zero = '0' then current_state <= 3;
                else current_state <= 9;
                end if;
      when 6 => current_state <= 7;
      when 7 => if mult_done = '1' then current_state <= 8; end if;
      when 8 => if count_equal_zero = '0' then current_state <= 3;
                else current_state <= 9;
                end if;
      when 9 => current_state <= 10;
      when 10 => if mult_done = '1' then current_state <= 11; end if;
      when 11 => current_state <= 12;
      when 12 => current_state <= 13;
      when 13 => current_state <= 14;
      when 14 => if mult_done = '1' then current_state <= 15; end if;
      when 15 => current_state <= 0;
    end case;
  end if;
  end process control_unit;

end circuit;


 